import { RdDUtility } from "./rdd-utility.js";
import { TMRUtility } from "./tmr-utility.js";
import { RdDRollDialogEthylisme } from "./rdd-roll-ethylisme.js";
import { RdDRoll } from "./rdd-roll.js";
import { RdDTMRDialog } from "./rdd-tmr-dialog.js";
import { Misc } from "./misc.js";
import { RdDAstrologieJoueur } from "./rdd-astrologie-joueur.js";
import { RdDResolutionTable } from "./rdd-resolution-table.js";
import { RdDDice } from "./rdd-dice.js";
import { RdDRollTables } from "./rdd-rolltables.js";
import { ChatUtility } from "./chat-utility.js";
import { RdDItemSort } from "./item-sort.js";
import { Grammar } from "./grammar.js";
import { RdDEncaisser } from "./rdd-roll-encaisser.js";
import { RdDCombat } from "./rdd-combat.js";

/* -------------------------------------------- */
/**
 * Extend the base Actor entity by defining a custom roll data structure which is ideal for the Simple system.
 * @extends {Actor}
 */
export class RdDActor extends Actor {

  /* -------------------------------------------- */
  /**
   * Override the create() function to provide additional RdD functionality.
   *
   * This overrided create() function adds initial items 
   * Namely: Basic skills, money, 
   *
   * @param {Object} data        Barebones actor data which this function adds onto.
   * @param {Object} options     (Unused) Additional options which customize the creation workflow.
   *
   */
   
  static async create(data, options) {
    
    // Case of compendium global import
    if (data instanceof Array) {
      return super.create(data, options);
    }
    // If the created actor has items (only applicable to duplicated actors) bypass the new actor creation logic
    if (data.items) {
      return super.create(data, options);
    }
    
    data.items = [];
    let compendiumName = "";
    if (data.type == "personnage") {
      compendiumName = "foundryvtt-reve-de-dragon.competences";
    } else if (data.type == "creature") {
      compendiumName = "foundryvtt-reve-de-dragon.competences-creatures";
    } else if (data.type == "entite") {
      compendiumName = "foundryvtt-reve-de-dragon.competences-entites";
    }
    let competences = [];
    const pack = game.packs.get(compendiumName);
    await pack.getIndex().then(index => competences = index);
    for (let comp of competences)
    {
      let compItem = undefined;
      await pack.getEntity(comp._id).then(skill => compItem = skill);
      data.items.push(compItem);
    }

    // Ajout monnaie
    if (data.type == "personnage" ) {
      await RdDActor.ajouterMonnaie( data.items );
    } 

    return super.create(data, options);
  }
  
  /* -------------------------------------------- */  
  prepareData() {
    super.prepareData();

    const actorData = this.data;

    // Dynamic computing fields
    this.encTotal = 0;

    /*
    // Auto-resize token
    if (this.isToken) {
      let tokenSize = actorData.data.carac.taille.value/10;
      this.token.update({height: tokenSize, width: tokenSize } );
    }*/
    
    // Make separate methods for each Actor type (character, npc, etc.) to keep
    // things organized.
    if (actorData.type === 'personnage') this._prepareCharacterData(actorData);
    if (actorData.type === 'creature') this.computeEtatGeneral(actorData);
    if (actorData.type === 'humanoide') this.computeEtatGeneral(actorData);
  }

  /* -------------------------------------------- */
  /**
   * Prepare Character type specific data
   */
  _prepareCharacterData(actorData) {
    // Initialize empty items
    RdDUtility.computeCarac(actorData.data);
    this.computeEncombrementTotalEtMalusArmure();
    this.computeEtatGeneral();
  }

    /* -------------------------------------------- */      
    static async ajouterMonnaie( items ) {
      let etain = { name: "Etain (1 denier)", type: 'monnaie', _id: randomID(),
                      data: { 
                        quantite: 0,
                        valeur_deniers: 1,
                        encombrement: 0.01,
                        description: ""
                        }
                      }
      items.push(etain);
  
      let bronze = duplicate(etain);
      bronze.id = randomID();
      bronze.name = "Bronze (10 deniers)";
      bronze.data.valeur_deniers = 10;
      items.push(bronze);
      
      let argent = duplicate(etain);
      argent.id = randomID();
      argent.name = "Argent (1 sol)";
      argent.data.valeur_deniers = 100;
      items.push(argent);
      
      let or = duplicate(argent);
      or.id = randomID();
      or.name = "Or (10 sols)";
      or.data.valeur_deniers = 1000;
      items.push(or);
    }
  
  /* -------------------------------------------- */
  isCreature() {
    return this.data.type == 'creature' || this.data.type == 'entite';
  }

  /* -------------------------------------------- */
  isPersonnage() {
    return this.data.type == 'personnage';
  }

  /* -------------------------------------------- */
  getReveActuel() {
    return this.data.data.reve.reve.value;
  }

  getChanceActuel() {
    return this.data.data.compteurs.chance.value;
  }

  /* -------------------------------------------- */
  getForceValue() {
    return this.data.data.carac.force ? this.data.data.carac.force.value : this.data.data.carac.reve.value;
  }

  /* -------------------------------------------- */
  getBonusDegat() {
    // TODO: gérer séparation et +dom créature/entité indépendament de la compétence
    return Misc.toInt(this.data.data.attributs.plusdom.value);
  }

  /* -------------------------------------------- */  
  getProtectionNaturelle() {
    return Misc.toInt(this.data.data.attributs.protection.value);
  }

  /* -------------------------------------------- */  
  getCompetenceList() {
    return this.data.items.filter( (item) => item.type == 'competence');
  }

  /* -------------------------------------------- */
  getCompetence(compName) {
    return RdDUtility.findCompetence(this.data.items, compName);
  }
  /* -------------------------------------------- */
  getBestDraconic() {
    const list = this.getDraconicList().sort((a, b) =>  b.data.niveau - a.data.niveau);
    if (list.length==0)
    {
      return { name: "none", niveau: -11 };
    }
    return duplicate(list[0]);
  }

  getEncombrementTotal() {
    return this.encTotal ? Math.floor(this.encTotal) : 0
  }
  /* -------------------------------------------- */
  async deleteSortReserve(sortReserve) {
    let reserve = duplicate(this.data.data.reve.reserve);
    let len = reserve.list.length;
    let i = 0;
    let newTable = [];
    for( i=0; i < len; i++) {
      if (reserve.list[i].coord != sortReserve.coord && reserve.list[i].sort.name != sortReserve.sort.name ) 
        newTable.push(reserve.list[i]);
    }
    if ( newTable.length != len ) {
      reserve.list = newTable;
      await this.update( {"data.reve.reserve": reserve } );    
    }
  }

  /* -------------------------------------------- */
  getSurprise() {
    if (this.isEntiteCauchemar()) {
      return '';
    }
    // TODO: gérer une liste de flags demi-surprise (avec icône sur le token)?
    if ( this.data.data.sante.sonne.value) {
      return 'demi';
    }
    return '';
  }

  /* -------------------------------------------- */
  isDemiSurprise() {
    return this.getSurprise() == 'demi';
  }

  /* -------------------------------------------- */
  isSurpriseTotale() {
    return this.getSurprise() == 'totale';
  }

  /* -------------------------------------------- */
  async dormirChateauDormant() {
    let message = { 
      whisper: ChatUtility.getWhisperRecipientsAndGMs( this.name ),
      content : ""
    };

    const blessures = duplicate(this.data.data.blessures);
    console.log("dormirChateauDormant", blessures)
    await this._recupererBlessures(message, "legere", blessures.legeres.liste.filter(b => b.active), []);
    await this._recupererBlessures(message, "grave", blessures.graves.liste.filter(b => b.active), blessures.legeres.liste);
    await this._recupererBlessures(message, "critique", blessures.critiques.liste.filter(b => b.active), blessures.graves.liste);
    await this.update( {"data.blessures": blessures } );
    await this._recupererVie(message);
    await this.transformerStress(message);
    await this.retourSeuilDeReve(message);
    message.content = "A la fin Chateau Dormant, " + message.content +"<br>Un nouveau jour se lève";
    ChatMessage.create( message );
  }

  /* -------------------------------------------- */
  async _recupererBlessures(message, type, liste, moindres) {
    let count = 0;
    const definitions = RdDUtility.getDefinitionsBlessures();
    let definition = definitions.find( d => d.type == type);
    for (let blessure of liste) {
      if (blessure.jours >= definition.facteur) {
        let rolled = await this._jetRecuperationConstitution(Misc.toInt(blessure.soins_complets), message);
        blessure.soins_complets = 0;
        if (rolled.isSuccess && this._retrograderBlessure(type, blessure, moindres)) {
            message.content += " -- une blessure " + type + " cicatrise";
            count++;
        }
        else if (rolled.isETotal) {
            message.content += " -- une blessure " + type + " s'infecte (temps de guérison augmenté de " + definition.facteur + " jours, perte de vie)";
            blessure.jours = 0;
            await this.santeIncDec("vie", -1);
        }
        else {
          message.content += " -- une blessure " + type + " reste stable";
        }
      }
      else {
        blessure.jours++;
      }
    }
  }

  /* -------------------------------------------- */
  _retrograderBlessure(type, blessure, blessuresMoindres)
  {
    if (type != "legere") {
      let retrograde = blessuresMoindres.find(b => !b.active);
      if (!retrograde) {
        return false;
      }
      mergeObject(retrograde, { "active": true, "premiers_soins": 0, "soins_complets": 0, "jours": 0, "loc": blessure.loc });
    }
    this._supprimerBlessure(blessure);
    return true;
  }
  
  /* -------------------------------------------- */
  _supprimerBlessure(blessure) {
    mergeObject(blessure, { "active": false, "premiers_soins": 0, "soins_complets": 0, "jours": 0, "loc": "" });
  }

  /* -------------------------------------------- */
  async _recupererVie(message) {
    let blessures = [].concat(this.data.data.blessures.legeres.liste).concat(this.data.data.blessures.graves.liste).concat(this.data.data.blessures.critiques.liste);
    let nbBlessures = blessures.filter(b => b.active);
    let vieManquante = this.data.data.sante.vie.max - this.data.data.sante.vie.value;
    if (nbBlessures == 0 && vieManquante>0) {
      let bonusSoins = 0;
      for (let b of blessures)
      {
        bonusSoins = Math.max(bonusSoins, Misc.toInt(b.soins_complets));
      }
      let rolled = await this._jetRecuperationConstitution(bonusSoins, message)
      if (rolled.isSuccess) {
        const gain = Math.min(rolled.isPart ? 2 : 1, vieManquante);
        message.content += " -- récupération de vie: " + gain; 
        await this.santeIncDec("vie", gain);
      }
      else if (rolled.isETotal) {
        message.content += " -- perte de vie: 1";
        await this.santeIncDec("vie", -1);
      }
      else{
        message.content +=" -- vie stationnaire ";
      }
    }
  }

  /* -------------------------------------------- */
  async _jetRecuperationConstitution(bonusSoins, message = undefined) {
    let difficulte = Misc.toInt(bonusSoins) + Math.min(0, this.data.data.sante.vie.value - this.data.data.sante.vie.max);
    let rolled = await RdDResolutionTable.roll(this.data.data.carac.constitution.value, difficulte);
    if (message) {
      message.content += RdDResolutionTable.explain(rolled).replace(/Jet :/, "Constitution :");
    }
    return rolled;
  }

  /* -------------------------------------------- */
  async remiseANeuf() {
    let message = { 
      whisper: ChatUtility.getWhisperRecipientsAndGMs( this.name ),
      content : "Remise à neuf de " + this.name
    };
    if (this.isEntiteCauchemar()) {
      await this.santeIncDec("endurance", this.data.data.sante.endurance.max - this.data.data.sante.endurance.value);
    }
    else {

      if (this.data.data.blessures){
        const blessures = duplicate(this.data.data.blessures);
        for (let listeBlessures of [blessures.legeres.liste, blessures.graves.liste, blessures.critiques.liste]) {
          for (let blessure of listeBlessures) {
            this._supprimerBlessure(blessure);
          }
        }
        await this.update( {"data.blessures": blessures } );
      }
      await this.setEthylisme(1);

      await this.santeIncDec("vie", this.data.data.sante.vie.max - this.data.data.sante.vie.value);
      await this.santeIncDec("endurance", this.data.data.sante.endurance.max - this.data.data.sante.endurance.value);
      if (this.data.data.sante.fatigue){
        let fatigue = duplicate(this.data.data.sante.fatigue)
        fatigue.value = 0;
        await this.update( {"data.sante.fatigue": fatigue } );
      }
    }
    ChatMessage.create( message );
  }


  /* -------------------------------------------- */
  async dormir(heures=1)  {
    let message = { 
      whisper: ChatUtility.getWhisperRecipientsAndGMs( this.name ),
      content : "Vous dormez " + heures + " heure" + (heures > 1 ? "s":  "")
    };
    await this.recupereEndurance(message);
    for (let i=0; i<heures; i++) {
      await this._recupererEthylisme(message);
      await this.recupererFatigue(message);
      await this.recuperationReve(message);
    }
    ChatMessage.create( message );
  }

  async _recupererEthylisme(message) {
    let ethylisme = duplicate(this.data.data.compteurs.ethylisme);
    ethylisme.nb_doses = 0;
    ethylisme.jet_moral = false;
    if (ethylisme.value < 1)
    {
      ethylisme.value = Math.min(ethylisme.value + 1, 1);
      if (ethylisme.value <= 0) {
        message.content += `<br>Vous dégrisez un peu (${RdDUtility.getNomEthylisme(ethylisme.value)})`;
      }
    }
    await this.update({ "data.compteurs.ethylisme": ethylisme });
  }

  /* -------------------------------------------- */
  async recupereEndurance(message) {
    const manquant = this._computeEnduranceMax() - this.data.data.sante.endurance.value;
    if (manquant > 0) {
      await this.santeIncDec("endurance", manquant);
      message.content += "<br>Vous récuperez " + manquant + " points d'endurance";
    }
  }

  /* -------------------------------------------- */
  async recupererFatigue(message) {
    let fatigue = duplicate(this.data.data.sante.fatigue)
    const fatigueMin = this._computeFatigueMin();
    if (fatigue.value <= fatigueMin) {
      message.content += "<br>Vous êtes déjà reposé";
      return;
    }
    fatigue.value = Math.max(fatigueMin, this._calculRecuperationSegment(fatigue.value));
    console.log("recupererFatigue", fatigue)
    await this.update( {"data.sante.fatigue": fatigue } );
    if (fatigue.value == 0)
    {
      message.content += "<br>Vous êtes bien reposé";
    }
  }

  /* -------------------------------------------- */
  _calculRecuperationSegment(actuel)
  {
    const segments = RdDUtility.getSegmentsFatigue(this.data.data.sante.endurance.max);
    let cumul = 0;
    let i;
    for (i=0; i <11; i++) {
      cumul += segments[i];
      let diff = cumul - actuel;
      if (diff >= 0)
      {
        const limit2Segments = Math.floor(segments[i] / 2);
        if (diff > limit2Segments && i > 0) {
          cumul -= segments[i-1]; // le segment est à moins de la moitié, il est récupéré 
        }
        cumul -= segments[i];
        break;
      }
    };
    return cumul;
  }

  /* -------------------------------------------- */
  async recuperationReve(message)  {
    const seuil = this.data.data.reve.seuil.value;
    const reveActuel = this.getReveActuel();
    if (reveActuel >= seuil) {
      message.content += "<br>Vous avez suffisament rêvé (seuil " + seuil + ", rêve actuel "+reveActuel+")";
    }
    else {
      let deRecuperation = await RdDDice.deDraconique();
      console.log("recuperationReve", deRecuperation);
      if (deRecuperation>=7)
      {
        // Rêve de Dragon !
        message.content += "<br>Vous faites un <strong>Rêve de Dragon</strong> de " + deRecuperation + " Points de rêve";
        message.content += await this.combattreReveDeDragon(deRecuperation);
      }
      else{
        message.content += "<br>Vous récupérez " + deRecuperation + " Points de rêve";
        await this.reveActuelIncDec(deRecuperation);
      }
    }
  }

  /* -------------------------------------------- */
  async retourSeuilDeReve(message) {
    const seuil = this.data.data.reve.seuil.value;
    const reveActuel = this.getReveActuel();
    if (reveActuel > seuil) {
      message.content += "<br>Votre rêve redescend vers son seuil naturel (seuil " + seuil + ", nouveau rêve actuel "+(reveActuel-1)+")";
      await this.reveActuelIncDec(-1);
    }
  }

  /* -------------------------------------------- */
  async combattreReveDeDragon(force){
    let draconic = this.getBestDraconic();
    let niveau = Math.max(0, draconic.data.niveau);
    let etat = this.data.data.compteurs.etat.value;
    let difficulte = niveau - etat - force;
    let reveActuel = this.getReveActuel();
    let rolled = await RdDResolutionTable.roll(reveActuel, difficulte);
    // TODO: xp particulière
    console.log("combattreReveDeDragon", rolled );
    return await this.appliquerReveDeDragon(rolled, force);
  }

  /* -------------------------------------------- */
  async appliquerReveDeDragon(roll, force) {
    let message = "";
    if (roll.isSuccess) {
      message += "<br>Vous gagnez " + force + " points de Rêve";
      await this.updatePointDeSeuil();
      await this.reveActuelIncDec(force);
    }
    if (roll.isPart) {
      // TODO: Dialog pour choix entre HR opu général?
      let tete = "à déterminer";
      message += "<br>Vous gagnez une Tête de dragon: " + tete;
    }
    if (roll.isEchec) {
      message += "<br>Vous subissez une Queue de Dragon: " + await this.ajouterQueue();
    }
    if (roll.isETotal) {
      message += "<br>A cause de votre échec total, vous subissez une deuxième Queue de Dragon: " + await this.ajouterQueue();
    }
    return message;
  }

  /* -------------------------------------------- */
  async sortMisEnReserve(rollData, sort) {
    let reserve = duplicate(this.data.data.reve.reserve);
    reserve.list.push({ coord: rollData.coord, sort: sort, draconic: duplicate(rollData.selectedDraconic) });
    await this.update({ "data.reve.reserve": reserve });
    this.currentTMR.updateSortReserve();
  }

  /* -------------------------------------------- */  
  updateCarac( caracName, caracValue )  
  {
    let caracpath = "data.carac." + caracName + ".value"
    if (caracName == "reve") {
      if (caracValue > Misc.toInt(this.data.data.reve.seuil.value)) {
        this.setPointsDeSeuil(caracValue);
      }
    }
    this.update( { caracpath: caracValue } );
  }
  
  /* -------------------------------------------- */  
  async updateCreatureCompetence( compName, fieldName, compValue )
  {
    let comp = this.getCompetence(compName);
    console.log( comp );
    if ( comp ) {
      const update = {_id: comp._id }
      if (fieldName == "niveau")
        update['data.niveau'] = compValue;
      else if (fieldName == "dommages")
        update['data.dommages'] = compValue;
      else 
        update['data.carac_value'] = compValue;
      console.log(update);
      const updated = await this.updateEmbeddedEntity("OwnedItem", update); // Updates one EmbeddedEntity
    }
  }

  /* -------------------------------------------- */  
  async updateCompetence( compName, compValue )
  {
    let comp = this.getCompetence(compName);
    if ( comp ) {
      let troncList = RdDUtility.isTronc( compName );
      let maxNiveau = compValue;
      if ( troncList  ) {
        let message = "Vous avez modifié une compétence 'tronc'. Vérifiez que les compétences suivantes évoluent ensemble jusqu'au niveau 0 : ";
        for(let troncName of troncList) {
          message += "<br>" + troncName;
        }
        ChatMessage.create( {
        whisper: ChatMessage.getWhisperRecipients(game.user.name),
        content: message } );
      }
      const update = {_id: comp._id, 'data.niveau': maxNiveau };
      const updated = await this.updateEmbeddedEntity("OwnedItem", update); // Updates one EmbeddedEntity
    } else {
      console.log("Competence not found", compName);
    }
  }
  
  /* -------------------------------------------- */  
  async updateCompetenceXP( compName, compValue )
  {
    let comp = this.getCompetence(compName);
    if ( comp ) {
      const update = {_id: comp._id, 'data.xp': compValue };
      const updated = await this.updateEmbeddedEntity("OwnedItem", update); // Updates one EmbeddedEntity
    } else {
      console.log("Competence not found", compName);
    }
  }
  
  /* -------------------------------------------- */  
  async updateCompteurValue( fieldName, fieldValue )
  { 
    //console.log("Update", fieldName, fieldValue);
    let compteurs = duplicate(this.data.data.compteurs);
    compteurs[fieldName].value =  fieldValue;
    await this.update( {"data.compteurs": compteurs } );
  }

  /* -------------------------------------------- */  
  async updateProtectionValue( fieldName, fieldValue )
  { 
    let attributs = duplicate(this.data.data.attributs);
    attributs[fieldName].value = fieldValue;
    await this.update( {"data.attributs": attributs } );
  }

  /* -------------------------------------------- */  
  /** Teste si le conteneur de destination a suffisament de capacité
   * pour recevoir le nouvel objet
   */
  testConteneurCapacite( itemId, conteneurId ) {
    if ( !conteneurId ) return true; // pas de conteneur (porté sur soi), donc toujours OK.
    let conteneur = this.items.find( conteneur => conteneurId == conteneur._id);  // recup conteneur
    //console.log("Conteneur trouvé : ", conteneur);
    if ( conteneur && conteneur.type == "conteneur" ) {
      let currentEnc = 0; // Calculer le total actuel des contenus
      for (let id of conteneur.data.data.contenu) {
        let objet = this.items.find( objet => (id == objet._id) );
        currentEnc += (objet) ? objet.data.data.encombrement : 0;
      } 
      // Et gérer le nouvel objet
      let nouvelObjet = this.items.find( objet => (itemId == objet._id) );
      if ( currentEnc + nouvelObjet.data.data.encombrement > Number(conteneur.data.data.capacite) )
        return false;
    }
    return true;
  }

  /* -------------------------------------------- */  
  /** Supprime un item d'un conteneur, sur la base 
   * de leurs ID */
  async enleverDeConteneur( itemId, conteneurId ) {
    if ( !conteneurId ) return; // pas de conteneur (porté sur soi)
    let conteneur = this.items.find( conteneur => conteneurId == conteneur._id);  // recup conteneur
    if ( conteneur ) { // Si présent
      let data2use = duplicate(conteneur.data);
      //console.log("Suppression du conteneur1", conteneurId, itemId,  conteneur.data.data.contenu);
      let contenu = data2use.data.contenu;
      let index = contenu.indexOf(itemId);
      while (index >= 0) { // Force cleanup, itemId is unique
        contenu.splice(index, 1);
        index = contenu.indexOf(itemId);
      }
      await this.updateEmbeddedEntity("OwnedItem", data2use); 
    }
  }

  /* -------------------------------------------- */  
  /** Ajoute un item dans un conteneur, sur la base 
   * de leurs ID */
  async ajouterAConteneur( itemId, conteneurId ) {
    if ( !conteneurId ) return; // pas de conteneur (porté sur soi)
    let conteneur = this.items.find( conteneur => conteneurId == conteneur._id);
    if ( conteneur && conteneur.type == 'conteneur' ) {
      let data2use = duplicate(conteneur.data);
      data2use.data.contenu.push( itemId );
      await this.updateEmbeddedEntity("OwnedItem", data2use );  
    }
  }

  /* -------------------------------------------- */  
  detectSurEncombrement( ) {
    let diffEnc = Number(this.encTotal) - Number(this.data.data.attributs.encombrement.value);
    if ( diffEnc > 0 ) { // Sur-encombrement
      let malus = Math.round( diffEnc);
      malus = (malus == 0) ? 1 : malus; // Always 1 at least
      //console.log("Sur enc malus", malus);
      return malus;
    }
    return 0;
  }

  /* -------------------------------------------- */  
  async computeEncombrementTotalEtMalusArmure( ) {
    let encTotal = 0;
    let malusArmureData = duplicate(this.data.data.attributs.malusarmure);
    let newMalusArmure  = 0;
    for (const item of this.data.items) {
      if ( item.type  == 'armure' && item.data.equipe ) {  // Armure équipée, intégration du malus armure total
        newMalusArmure += item.data.malus;
      }
      // Calcul encombrement
      if ( item.data && item.data.encombrement != undefined ) {
        if ( !Number(item.data.encombrement) ) item.data.encombrement = 0; // Auto-fix
        if ( item.data.quantite == undefined) item.data.quantite = 1; // Auto-fix
        if ( item.data.quantite < 0) item.data.quantite = 0; // Auto-fix
        item.data.encTotal = Number(item.data.encombrement) * Number(item.data.quantite);
        //console.log("Enc:", item.name, item.data.encombrement, item.data.quantite, item.data.encTotal);
        encTotal += item.data.encTotal;
      } else {
        item.data.encTotal = 0; // Force default enc
      }
    }
    // Mise à jour valeur totale et états
    this.encTotal = encTotal;
    this.detectSurEncombrement();
    // Mise à jour éventuelle du malus armure
    if (newMalusArmure != malusArmureData.value) {
      malusArmureData.value = newMalusArmure;
      await this.update( {"data.attributs.malusarmure": malusArmureData } );
    }
  }

  /* -------------------------------------------- */  
  computeResumeBlessure(blessures = this.data.data.blessures) {
    let nbLegeres   = this.countBlessures(blessures.legeres.liste );
    let nbGraves    = this.countBlessures(blessures.graves.liste );
    let nbCritiques = this.countBlessures(blessures.critiques.liste );

    let resume = "Blessures:";
    if (nbCritiques > 0 || nbGraves > 0 || nbLegeres > 0) {
      if (nbLegeres > 0) {
        resume += " " + nbLegeres + " légères";
      }
      if (nbGraves > 0) {
        if (nbLegeres > 0)
          resume += ",";
        resume += " " + nbGraves + " graves";
      }
      if (nbCritiques > 0) {
        if (nbGraves > 0 || nbLegeres > 0)
          resume += ",";
        resume += " une CRITIQUE !";
      }
    }
    else {
      resume += " aucune";
    }
    return resume;
  }

  /* -------------------------------------------- */  
  computeEtatGeneral( )
  {
    let data = this.data.data;
    // Pas d'état général pour les entités forçage à 0
    if ( this.data.type == 'entite') {
      data.compteurs.etat.value = 0;
      return;
    }
    // Pour les autres
    let state = - (data.sante.vie.max - data.sante.vie.value);
    if (data.sante.fatigue) // Creatures n'ont pas de fatigue
      state += RdDUtility.currentFatigueMalus(data.sante.fatigue.value, data.sante.endurance.max);
    if (data.compteurs && data.compteurs.ethylisme && data.compteurs.ethylisme.value < 0 ) { // Ajout de l'éthylisme
      state += data.compteurs.ethylisme.value;
    }
    data.compteurs.etat.value   = state;
    if ( data.compteurs && data.compteurs.surenc) {
      data.compteurs.surenc.value = - this.detectSurEncombrement();
    }
  }
  
  /* -------------------------------------------- */  
  async ajouterRefoulement( value=1) {
    let ret = "none";
    
    let refoulement = duplicate(this.data.data.reve.refoulement);
    refoulement.value = refoulement.value + value;

    let total = new Roll("1d20").roll().total;
    if ( total <= refoulement.value ) {
      refoulement.value = 0;
      this.ajouterSouffle({chat: true});
      ret = "souffle";
    }

    await this.update( {"data.reve.refoulement": refoulement } );
    return ret;      
  }

  /* -------------------------------------------- */
  async ajouterSouffle(options = {chat : false}) {
    let souffle = await RdDRollTables.getSouffle();
    await this.createOwnedItem(souffle);
    if (options.chat){
      ChatMessage.create({
        whisper: ChatUtility.getWhisperRecipientsAndGMs( this.name ),
        content: this.name + " subit un Souffle de Dragon  : " + souffle.name
      });
    }
    return souffle;
  }

  /* -------------------------------------------- */
  async ajouterQueue(options = {chat : false}) {
    // TODO: Déterminer si Thanatos a été utilisé? => laisser le joueur ne pas choisir Thanatos => choisir sa voie?
    let utiliseThanatos = false;
    let queue;
    if (utiliseThanatos) {
      queue = await RdDRollTables.getOmbre();
      // mettre à jour: plus d'ombre en vue
    }
    else {
      queue = await RdDRollTables.getQueue();
    }
    await this.createOwnedItem(queue);
    if (options.chat){
      ChatMessage.create({
        whisper: ChatUtility.getWhisperRecipientsAndGMs( this.name ),
        content: this.name + " subit une Queue de Dragon : " + queue.name
      });
    }
    return queue;
  }
  
  /* -------------------------------------------- */  
  displayTMRQueueSouffleInformation( ) {
    for ( let item of this.data.items) {
      let content
      if (item.type == 'queue') {
        if ( item.name.toLowerCase() == 'conquête' ) {
          content = "RAPPEL ! Vous souffrez d'une <strong>Conquête</strong> : " + item.data.description;
        } 
        else if ( item.name.toLowerCase() == 'pélerinage' ) {
          content = "RAPPEL ! Vous souffrez d'un <strong>Pélerinage</strong> : " + item.data.description;
        }
        else if ( item.name.toLowerCase() == 'urgence draconique' ) {
          content = "RAPPEL ! Vous souffrez d'une <strong>Urgence Draconique</strong> : " + item.data.description;
        }
      } else if ( item.type == 'souffle') {
        if ( item.name.toLowerCase() == 'périple' ) {
          content = "RAPPEL ! Vous souffrez du Souffle <strong>Périple</strong>. Vous devez gérer manuellement le détail du Périple.<br>" + item.data.description;
        } else if ( item.name.toLowerCase() == 'fermeture des cités' ) {
          content = "RAPPEL ! Vous souffrez du Souffle <strong>Fermeture des Cités</strong>. Vous devez gérer manuellement le détail des Citées ré-ouvertes.<br>" + item.data.description;
        } else if ( item.name.toLowerCase() == 'désorientation' ) {
          content = "RAPPEL ! Vous souffrez du Souffle <strong>Désorientation</strong>. Vous devez gérer avec votre MJ les effets de ce souffle.<br>" + item.data.description;
        } else if ( item.name.toLowerCase() == 'double résistance du fleuve' ) {
          content = "RAPPEL ! Vous souffrez du Souffle <strong>Double Résistance du Fleuve</strong>. Vous devez gérer avec votre MJ les effets de ce souffle.<br>" + item.data.description;
        }
      }
      if (content) {
        ChatMessage.create({
          whisper: ChatUtility.getWhisperRecipientsAndGMs( this.name ),
          content: content
        });
      }  
    }
  }

  /* -------------------------------------------- */  
  async deleteTMRRencontreAtPosition( ) {
    let rencontres = duplicate(this.data.data.reve.rencontre);
    let len = rencontres.list.length;
    let i = 0;
    //console.log("List", rencontres, len);
    let newTable = [];
    for( i=0; i < len; i++) {
      if (rencontres.list[i].coord != this.data.data.reve.tmrpos.coord ) 
        newTable.push(rencontres.list[i]);
    }
    if ( newTable.length != len ) {
      rencontres.list = newTable;
      //console.log("Result: ", rencontres);
      await this.update( {"data.reve.rencontre": rencontres } );    
    }
  }
  
  /* -------------------------------------------- */  
  async addTMRRencontre( currentRencontre ) {    
    let rencontres = duplicate(this.data.data.reve.rencontre);
    let len = rencontres.list.length;
    let i = 0;
    let already = false;
    for( i=0; i < len; i++) {
      if (rencontres.list[i].coord == this.data.data.reve.tmrpos.coord ) 
        already = true;
    }
    if ( !already ) {
      rencontres.list.push( {coord: this.data.data.reve.tmrpos.coord, rencontre: currentRencontre} ); 
      await this.update( {"data.reve.rencontre": rencontres } );
    }
  }

  /* -------------------------------------------- */  
  async updateCoordTMR( coord ) {        
    let tmrPos = duplicate(this.data.data.reve.tmrpos );
    tmrPos.coord = coord;
    await this.update( {"data.reve.tmrpos": tmrPos } );
  }
  
  /* -------------------------------------------- */  
  async reveActuelIncDec( value ) {
    let reve = duplicate(this.data.data.reve.reve);
    reve.value = Math.max(reve.value + value, 0);
    await this.update( {"data.reve.reve": reve } );
  }
  
  /* -------------------------------------------- */
  async updatePointDeSeuil(value=1) {
    const seuil = Misc.toInt(this.data.data.reve.seuil.value);
    const reve = Misc.toInt(this.data.data.carac.reve.value);
    if (seuil < reve) {
      await this.setPointsDeSeuil(Math.min(seuil+value, reve));
    }
  }
  
  /* -------------------------------------------- */
  async setPointsDeSeuil( value ) {    
    let seuil = duplicate(this.data.data.reve.seuil);
    seuil.value = value;
    await this.update( {"data.reve.seuil": seuil } );
  }

  /* -------------------------------------------- */  
  testSiSonne( sante, endurance ) 
  {
    let result = new Roll("1d20").roll().total;
    if ( result <= endurance)
       sante.sonne.value = false;
    if ( result > endurance || result == 20) // 20 is always a failure
       sante.sonne.value = true;
    if (result == 1) {
      sante.sonne.value = false;
      let xp  = Misc.toInt(this.data.data.carac.constitution.xp) + 1;
      this.update( {"data.carac.constitution.xp": xp } ); // +1 XP !
      // TODO : Output to chat
    }
  }

  /* -------------------------------------------- */  
  countBlessures( blessuresListe ) 
  {
    return blessuresListe.filter(b => b.active).length
  }
  
  /* -------------------------------------------- */  
  async jetVie() {
    let myRoll = new Roll("1d20").roll();
    myRoll.showDice = true;
    await RdDDice.show(myRoll);

    let msgText = "Jet de Vie : " + myRoll.total + " / " + this.data.data.sante.vie.value + "<br>";
    if ( myRoll.total <= this.data.data.sante.vie.value ) {
      msgText += "Jet réussi, pas de perte de point de vie (prochain jet dans 1 round pour 1 critique, SC minutes pour une grave)";
      if ( myRoll.total == 1) {
        msgText += "La durée entre 2 jets de vie est multipliée par 20 (20 rounds pour une critique, SCx20 minutes pour une grave)";
      }
    } else {
      msgText += "Jet échoué, vous perdez 1 point de vie";
      await this.santeIncDec("vie", -1);
      if ( myRoll.total == 20) {
        msgText += "Votre personnage est mort !!!!!";
      }
    }
    const message = {
      content: msgText,
      whisper: ChatMessage.getWhisperRecipients(game.user.name)
    };
    ChatMessage.create(message);    
  }

  /* -------------------------------------------- */  
  async santeIncDec(name, inc, isCritique = false) {
    const sante = duplicate(this.data.data.sante);
    let data = sante[name];
    if (data==undefined) {
      return;
    }
    let minValue = 0;
    if (this.type == 'personnage') {
      // TODO: les animaux/humanoïdes on théoriquement aussi un sconst, mais la SPA n'est pas passé par là
      minValue = name == "vie" ? -Number(this.data.data.attributs.sconst.value) : 0;
    }

    let newValue = Math.max(minValue, Math.min(data.value + inc, data.max));
    //console.log("New value ", inc, minValue, newValue);

    if (name == "endurance" && this.data.type != 'entite' ) {
      if ( sante.fatigue && inc < 0 ) { // Each endurance lost -> fatigue lost
        sante.fatigue.value = sante.fatigue.value - inc
      }
      if ( !isCritique &&  newValue == 0 && inc < 0) { // perte endurance et endurance devient 0 -> -1 vie sauf si coup critique
        sante.vie.value = sante.vie.value - 1;
      }
      newValue = Math.max(0, newValue);
      if (inc>0) { // le max d'endurance s'applique seulement à la récupération
        newValue = Math.min(newValue, this._computeEnduranceMax())
      }
      if (data.value - newValue > 1) {
        this.testSiSonne(sante, newValue); // Peut-être sonné si 2 points d'endurance perdus d'un coup
      } else if (inc>0) {
        sante.sonne.value = false;
      }
    }
    data.value = newValue;
    //console.log(name, inc, data.value, newValue, minValue, data.max);
    if ( sante.fatigue) { // If endurance lost, then the same amount of fatigue cannot be recovered
      sante.fatigue.value = Math.max(sante.fatigue.value, this._computeFatigueMin());
    } 
    //console.log("SANTE::::", sante);
    
    await this.update( {"data.sante": sante } );
  }

  /* -------------------------------------------- */  
  _computeFatigueMin() {
    return this.data.data.sante.endurance.max - this.data.data.sante.endurance.value;
  }
  
  /* -------------------------------------------- */  
  _computeEnduranceMax() {
    let blessures = this.data.data.blessures;
    let diffVie = this.data.data.sante.vie.max - this.data.data.sante.vie.value;
    let maxEndVie = this.data.data.sante.endurance.max - (diffVie * 2);
    let nbGraves = this.countBlessures(blessures.graves.liste);
    let nbCritiques = this.countBlessures(blessures.critiques.liste);
    let maxEndGraves = Math.floor(this.data.data.sante.endurance.max / (2 * nbGraves));
    let maxEndCritiques = nbCritiques > 0 ? 1 : this.data.data.sante.endurance.max;
    return Math.max(0, Math.min(maxEndVie, maxEndGraves, maxEndCritiques));
  }

  /* -------------------------------------------- */  
  async manageBlessureFromSheet( bType, index, active ) {
    let bList    = duplicate(this.data.data.blessures);
    let blessure = bList[bType+"s"].liste[index];
    blessure.active = !blessure.active;
    if ( !blessure.active ) {
      blessure.premiers_soins = 0;
      blessure.soins_complets = 0;
      blessure.jours = 0;
      blessure.loc = "";  
    }
    //console.log("Blessure update", bType, index, blessure, bList );
    await this.update( { 'data.blessures': bList } );
  }

  /* -------------------------------------------- */  
  async setDataBlessureFromSheet( bType, index, psoins, pcomplets, jours, loc) {
    let bList    = duplicate(this.data.data.blessures);
    let blessure = bList[bType+"s"].liste[index];
    blessure.premiers_soins = psoins;
    blessure.soins_complets = pcomplets;
    blessure.jours = jours;
    blessure.loc = loc;
    await this.update( { 'data.blessures': bList } );    
  }

  /* -------------------------------------------- */  
  manageBlessures( blessuresData ) 
  {
    // Fast exit
    if ( this.data.type == 'entite') return; // Une entité n'a pas de blessures
    if ( blessuresData.legeres + blessuresData.graves + blessuresData.critiques == 0 ) return;

    let workData = duplicate(blessuresData);    
    let blessures = duplicate(this.data.data.blessures);
    // Manage blessures
    if ( workData.legeres > 0 ) { 
      for (let k=0; k<blessures.legeres.liste.length; k++) {
        let bless = blessures.legeres.liste[k];
        if ( !bless.active ){
          bless.active = true;
          bless.loc = workData.locName;
          workData.legeres--;
        }
        if (workData.legeres == 0) break;
      }
    }      
    
    if ( workData.legeres > 0 ) {
      workData.graves += 1;
      blessuresData.graves += 1;
    }
      
    if ( workData.graves > 0) {
      for (let k=0; k<blessures.graves.liste.length; k++) {
        let bless = blessures.graves.liste[k];
        if ( !bless.active ) {
          bless.active = true;
          bless.loc = workData.locName;
          workData.graves--;
        }
        if ( workData.graves == 0) break;
      }
    }

    if ( workData.graves > 0 ) { 
      workData.critiques = 1;
      blessuresData.critiques = 1;
    }
      
    if ( workData.critiques > 0 ) {
      workData.endurance = this.data.data.sante.endurance.value; // Patch with real endurance current value (ie end -> 0 when critique)
      blessures.critiques.liste[0].active = true;
      blessures.critiques.liste[0].loc  = workData.locName;
    }
      
    this.update( { "data.blessures": blessures } );
  }
  
  /* -------------------------------------------- */  
  async jetDeMoral(situation) {
    let jetMoral = new Roll("1d20").roll();
    RdDDice.show(jetMoral);
    let moralActuel = Misc.toInt(this.data.data.compteurs.moral.value);
    const difficulte = 10 + moralActuel;
    const succes = jetMoral.total <= difficulte;

    let ajustementMoral = this._calculAjustementMoral(succes, moralActuel, situation);
    
    await this.moralIncDec(ajustementMoral);
    ChatMessage.create({
      whisper: ChatUtility.getWhisperRecipientsAndGMs(game.user.name),
      content: `Jet de moral ${succes? "réussi": "manqué"} en situation ${situation} (${jetMoral.total}/${difficulte}), vous ${ajustementMoral>0? "gagnez du moral" : ajustementMoral<0? "perdez du moral": "gardez votre moral"}`
    });
    return ajustementMoral;
  }

  async moralIncDec(ajustementMoral) {
    let compteurs = duplicate(this.data.data.compteurs);
    compteurs.moral.value = Misc.toInt(compteurs.moral.value);;

    if (ajustementMoral != 0) {
      compteurs.moral.value += ajustementMoral;
      if (compteurs.moral.value > 3) {
        // exaltation
        compteurs.moral.value--;
        compteurs.exaltation.value = Misc.toInt(compteurs.exaltation.value) + 1;
      }
      if (compteurs.moral.value < -3) {
        // dissolution
        compteurs.moral.value++;
        compteurs.dissolution.value = Misc.toInt(compteurs.dissolution.value) + 1;
      }
      await this.update({ 'data.compteurs': compteurs });
    }
  }

  /* -------------------------------------------- */  
  _calculAjustementMoral(succes, moral, situation) {
    switch (situation) {
      case 'heureuse': return succes ? 1 : 0;
      case 'malheureuse': return succes ? 0 : -1;
      case 'neutre':
        if (succes && moral <= 0) return 1;
        if (!succes && moral > 0) return -1;
    }
    return 0;

  }
  async setEthylisme(degre) {
    let ethylisme = duplicate(this.data.data.compteurs.ethylisme);
    ethylisme.value = degre;
    ethylisme.nb_doses = 0;
    if (degre == 1) {
      ethylisme.jet_moral = false;
    }
    await this.update({ "data.compteurs.ethylisme": ethylisme});
  }

  /* -------------------------------------------- */
  async ethylismeTest() {
    let rollData = { 
      vieValue: this.data.data.sante.vie.value,
      etat: this.data.data.compteurs.etat.value - Math.min(0, this.data.data.compteurs.ethylisme.value), // Pour les jets d'Ethylisme, on ignore le degré d'éthylisme (p.162)
      diffNbDoses: -Number(this.data.data.compteurs.ethylisme.nb_doses || 0),
      finalLevel: 0,
      diffConditions: 0,
      ajustementsConditions: CONFIG.RDD.ajustementsConditions,
      forceAlcool: 0
    }
    let html = await renderTemplate('systems/foundryvtt-reve-de-dragon/templates/dialog-roll-ethylisme.html', rollData);
    new RdDRollDialogEthylisme(html, rollData, this ).render(true);
  }
  
  /* -------------------------------------------- */
  async performEthylisme( rollData ) {
    let ethylisme = duplicate(this.data.data.compteurs.ethylisme);
    
    // Je d'ethylisme
    let rollEthylisme = await RdDResolutionTable.roll( rollData.vieValue, rollData.finalLevel);    
    let msgText = RdDResolutionTable.explain(rollEthylisme) + "<br>";
    if (rollEthylisme.isSuccess ) {
      ethylisme.nb_doses ++;
      msgText += `Vous avez réussi votre jet d'éthylisme, vous avez consommé ${ethylisme.nb_doses} doses sans effet.`;
    } else {
      ethylisme.value = Math.max(ethylisme.value -1, -7);
      ethylisme.nb_doses = 0;

      let enduranceLost = new Roll("1d6").roll().total;
      await this.santeIncDec("endurance", -enduranceLost);
      const ajustementEthylique = ethylisme.value;
      // Qui a bu boira (p 164)
      let rollVolonte = await RdDResolutionTable.roll( this.data.data.carac.volonte.value, Math.min(ajustementEthylique, 0) + this.data.data.compteurs.moral.value); 
      msgText += `Vous avez échoué à votre jet d'éthylisme, vous êtes maintenant ${RdDUtility.getNomEthylisme(ajustementEthylique)} (${ajustementEthylique}).`
      msgText += "<br>" + RdDResolutionTable.explain(rollVolonte) + "<br>";
      msgText += "Qui a bu boira : " +(rollVolonte.isSuccess
        ? "vous êtes libre de continuer à boire ou pas."
        : "vous avez une envie irrépréssible de reprendre un verre.");
    }

    ChatMessage.create({
      content: msgText,
      whisper: ChatUtility.getWhisperRecipientsAndGMs(game.user.name)
    });    
    
    if (rollEthylisme.isEchec) {
      await this._jetDeMoralEthylique(ethylisme);
    }
    await this.update( { 'data.compteurs.ethylisme': ethylisme} );
  }
    
  /* -------------------------------------------- */
  async _jetDeMoralEthylique(ethylisme) {
    if (ethylisme.value >= -1 && !ethylisme.jet_moral) {
      let adjust = await this.jetDeMoral('heureuse');
      if (adjust > 0 || ethylisme.value == -1) {
        ethylisme.jet_moral = true;
      }
      if (ethylisme.value == -1 && adjust <= 0) {
        // alcool triste
        ChatMessage.create({
          content: "Décidément, vous avez l'alcool triste, vous perdez finalement un point de moral!",
          whisper: ChatUtility.getWhisperRecipientsAndGMs(game.user.name)
        });    
        this.moralIncDec(-1);
      }
    }
  }

  /* -------------------------------------------- */
  async stressTest() {
    const message = {
      content: "",
      whisper: ChatUtility.getWhisperRecipientsAndGMs(game.user.name)
    };
    await this.transformerStress(message);

    ChatMessage.create(message);
  }

  /* -------------------------------------------- */
  async transformerStress(message) {
    let compteurs = duplicate(this.data.data.compteurs);
    const stress = Misc.toInt(compteurs.stress.value);
    
    if (stress<=0) {
      return;
    }
    
    
    let stressRoll = await this._stressRoll();
    let convertis = Math.floor(stress * stressRoll.factor);
    compteurs.stress.value = Math.max(stress - convertis - 1, 0);
    
    let dissolution = Math.max(0, Misc.toInt(compteurs.dissolution.value));
    let exaltation = Math.max(0, Misc.toInt(compteurs.exaltation.value));
    const annule = Math.min(dissolution, exaltation);
    dissolution -= annule;
    exaltation -= annule;
    if (dissolution>0) {
      const perdus = Math.min(dissolution, convertis);
      convertis -= perdus;
      dissolution -= perdus;
    }
    compteurs.experience.value += convertis + exaltation;
    compteurs.dissolution.value = dissolution;
    compteurs.exaltation.value = 0;
    message.content += "<br>Vous transformez " + convertis + " points de Stress en Expérience" + stressRoll.comment;
    await this.update({ "data.compteurs": compteurs });
  }

  /* -------------------------------------------- */
  async _stressRoll() {
    let reveActuel = this.getReveActuel();
    let result = await RdDResolutionTable.roll(reveActuel, 0);
    console.log("_stressRoll", result);
    switch (result.code) {
      case "sign": return { factor: 0.75, comment: " (75%): " + result.quality + " - " + result.roll + " sur " + result.score + "%" }
      case "norm": return { factor: 0.5, comment: " (50%): " + result.quality + " - " + result.roll + " sur " + result.score + "%" }
      case "echec": return { factor: 0.2, comment: " (20%): " + result.quality + " - " + result.roll + " sur " + result.score + "%" }
      case "epart": return { factor: 0.1, comment: " (10%): " + result.quality + " - " + result.roll + " sur " + result.score + "%" }
      case "etotal": return { factor: 0, comment: " (0%): " + result.quality + " - " + result.roll + " sur " + result.score + "%" }
      case "part":
        {
          let second = await RdDResolutionTable.roll(reveActuel, 0);
          console.log("_stressRoll", second);
          switch (second.code) {
            case "part": case "sign":
              return { factor: 1.5, comment: " (150%): Double Particulière - " + result.roll + " puis " + second.roll + " sur " + result.score + "%" }
            default:
              return { factor: 1, comment: " (150%): " + result.quality + " - " + result.roll + " puis " + second.roll + " sur " + result.score + "%" }
          }
        }
    }
  }

  /* -------------------------------------------- */  
  createCallbackExperience() {
    return {
      condition: r => r.rolled.isPart && r.finalLevel < 0 && game.settings.get("core", "rollMode") != 'selfroll',
      action: r => this._appliquerAjoutExperience(r)
    };
  }

  /* -------------------------------------------- */  
  async _appliquerAjoutExperience(rollData) { 
    let xpResult = this.appliquerExperience( rollData.rolled, rollData.selectedCarac.label, (rollData.competence) ? rollData.competence.data.name: undefined );
    if (xpResult.result ) { 
      let xpmsg = "<br>Points d'expérience gagnés ! Carac: " + xpResult.xpCarac + ", Comp: " + xpResult.xpCompetence;
      let message = ChatUtility.prepareChatMessage('gmroll', this.name);
      message.content = "<strong>" + rollData.selectedCarac.label + "</strong>"
            + xpmsg;
      ChatMessage.create(message);
    }
  }

  /* -------------------------------------------- */
  async rollUnSort(coord) {
    let sortList = duplicate(this.getSortList()); // Duplication car les pts de reve sont modifiés dans le sort
    if (!sortList || sortList.length == 0)
    {
      ui.notifications.info("Aucun sort disponible!");
      return;
    }
    
    let rollData = { 
      selectedCarac: this.data.data.carac.reve,
      draconicList: this.getDraconicList(),
      sortList: sortList,
      selectedDraconic: this.getBestDraconic(),
      selectedSort: sortList[0],
      coord: coord,
      coordLabel: TMRUtility.getTMRDescription( coord).label,
      diffLibre: sortList[0].data.difficulte, // Per default at startup
      coutreve: Array(20).fill().map((item, index) => 1 + index)
    }

    if ( this.currentTMR) this.currentTMR.minimize(); // Hide
    const dialog = await RdDRoll.create(this, rollData,
      { html: 'systems/foundryvtt-reve-de-dragon/templates/dialog-roll-sort.html', 
        close: html => { this.currentTMR.maximize() } // Re-display TMR
      },
      {
        name: 'lancer-un-sort',
        label: 'Lancer un sort',
        callbacks: [
          this.createCallbackExperience(),
          { action: r => this._rollUnSortResult(r, false) }
        ]
      },
      {
        name: 'mettre-en-reserve',
        label: 'Mettre un sort en réserve',
        callbacks: [
          this.createCallbackExperience(),
          { action:  r => this._rollUnSortResult(r, true) }
        ]
      }
    );
    dialog.render(true);
  }

  /* -------------------------------------------- */
  isRencontreSpeciale( ) { // Gestion queue/souffle 'Mauvaise Rencontre en Perpective'
    let addMsg = "";
    let rencSpecial = this.data.items.find( item => (item.type == 'queue' || item.type == 'souffle') && item.name.toLowerCase().includes( 'mauvaise rencontre' ) );
    if ( rencSpecial ) {
      rencSpecial = duplicate(rencSpecial); // To keep it
      if ( rencSpecial.type == 'queue') {
        this.deleteOwnedItem( rencSpecial._id ); // Suppression dans la liste des queues
        addMsg = " La queue a été supprimée de votre fiche automatiquement";
      } else {
        addMsg = " Vous devez gérer manuellement le décompte de mauvaises rencontres manuellement.";
      }        
      ChatMessage.create( {
        content: "Vous êtes sous le coup d'une Mauvaise Rencontre en Persective." + addMsg,
        whisper: ChatMessage.getWhisperRecipients(game.user.name) } );
    }
    return rencSpecial;
  }

  /* -------------------------------------------- */
  getTMRFatigue() { // Pour l'instant uniquement Inertie Draconique
    let inertieDraconique = this.data.items.find( item => item.type == 'queue' && item.name.toLowerCase().includes( 'inertie draconique' ) );
    if ( inertieDraconique ) {
      ChatMessage.create( {
        content: "Vous êtes sous le coup d'une Inertie Draconique : vous perdez 2 cases de Fatigue par déplacement au lieu d'1.",
        whisper: ChatMessage.getWhisperRecipients(game.user.name) } );
      return 2;
    }
    return 1;
  }

  /* -------------------------------------------- */
  isConnaissanceFleuve( ) { 
    return this.data.items.find( item => item.type == 'tete' && item.name.toLowerCase().includes( 'connaissance du fleuve' ) );
  }

  /* -------------------------------------------- */
  isReserveEnSecurite() {
    let reserveSecurite = this.data.items.find( item => item.type == 'tete' && item.name.toLowerCase().includes( ' en sécurité' ) );
    return reserveSecurite;
  }

  /* -------------------------------------------- */
  isDoubleResistanceFleuve( ) {
    let resistFleuve = this.data.items.find( item => item.type == 'souffle' && item.name.toLowerCase().includes( 'résistance du fleuve' ) );
    if ( resistFleuve ) {
      ChatMessage.create( {
        content: "Vous êtes sous le coup d'une Double Résistance du Fleuve : vous devez maîtriser 2 fois chaque case humide, un second jet est donc effectué.",
        whisper: ChatMessage.getWhisperRecipients(game.user.name) } );
      return true;
    }
    return false;
  }

  /* -------------------------------------------- */
  async checkSoufflePeage(cellDescr) {
    let peage = this.data.items.find( item => item.type == 'souffle' && item.name.toLowerCase().includes( 'péage' ) );
    if (peage && (cellDescr.type == 'pont' || cellDescr.type == 'cite') ) {
      await this.reveActuelIncDec( -1 );
      ChatMessage.create( {
        content: "Vous êtes sous le coup d'un Péage : l'entrée sur cette case vous coûte 1 Point de Rêve (déduit automatiquement).",
        whisper: ChatMessage.getWhisperRecipients(game.user.name) } );
    }
  }

  /* -------------------------------------------- */
  checkTeteDeplacementAccelere() {
    let deplAccelere = this.data.items.find( item => item.type == 'tete' && item.name.toLowerCase().includes( ' déplacement accéléré' ) );
    if ( deplAccelere ) {
      return true;
    }
    return false;
  }

  /* -------------------------------------------- */
  checkIsAdditionnalHumide( cellDescr, coordTMR) {
    let pontHumide = this.data.items.find( item => item.type == 'souffle' && item.name.toLowerCase().includes( ' des ponts' ) );
    if ( pontHumide && cellDescr.type == 'pont') {
      ChatMessage.create( {
        content: "Vous êtes sous le coup d'une Impraticabilité des Ponts : ils doivent être maîtrisés comme des cases humides.",
        whisper: ChatMessage.getWhisperRecipients(game.user.name) } );
      return true;
    }
    // Débordement ?
    let debordementList = this.data.items.filter( item => item.type == 'casetmr' && item.data.specific == 'debordement' );
    for (let caseTMR of debordementList) {
      if ( caseTMR.data.coord == coordTMR )
        return true;
    }
    
    return false;
  }

  /* -------------------------------------------- */
  async _rollUnSortResult(rollData, isSortReserve = false) {
    rollData.isSortReserve = isSortReserve;
    let rolled = rollData.rolled;
    let sort = rollData.selectedSort;
    let closeTMR = !rollData.isSortReserve;
    if (sort.data.isrituel && isSortReserve) {
      ui.notifications.error("Impossible de mettre le rituel '" + sort.name + "' en réserve");
      this.currentTMR.close(); // Close TMR !
      return;
    }

    let explications = rollData.isSortReserve
    ? ("<br>Mise en réserve du sort en " + rollData.coordLabel + "(" + rollData.coord + ")<strong>")
    : "<br>Lancement du sort <strong>";
    explications += sort.name + "</strong> : " + Misc.upperFirst(sort.data.draconic)
    + " pour " + sort.data.ptreve_reel + " points de Rêve"
    + "<br>Depuis la case " + rollData.coord + " (" + TMRUtility.getTMRDescription(rollData.coord).label + ")";
    
    let depenseReve = Number(sort.data.ptreve_reel);
    let myReve = duplicate(this.data.data.reve.reve);
    if (rolled.isSuccess) { // Réussite du sort !
      //sort.ptreve_reel = coutReve;
      if (rolled.isPart) {
        depenseReve = Math.max(Math.floor(depenseReve / 2), 1);
      }
      if (rollData.isSortReserve) {
        depenseReve++;
      }

      if (myReve.value > depenseReve) {
        explications += "<br>Réussite du sort: " + depenseReve + " points de Rêve sont dépensés (Bonus de case en " + rollData.coord + ": +" + rolled.bonus + "%)";

        // Incrémenter/gére le bonus de case
        RdDItemSort.incrementBonusCase(this, sort, rollData.coord);

        if (rollData.isSortReserve) {
          await this.sortMisEnReserve(rollData, sort);
          closeTMR = false;
        }
      }
      else {
        // Todo 0 pts de reve !!!!
        depenseReve = 0;
        explications += "<br>Pas assez de rêve";
        mergeObject(rollData, RdDResolutionTable.getResultat("echec"));
      }
    } else {
      if (rolled.isETotal) { // Echec total !
        depenseReve = Math.min(myReve.value, Math.floor(depenseReve * 1.5));
        explications += "<br><strong>Echec TOTAL</strong> du sort : " + depenseReve + " Points de Rêve";
        // TODO: mise en réserve d'un échec total...
      } else {
        depenseReve = 0
        explications += "<br>Echec du sort !";
      }
    }

    myReve.value = Math.max(myReve.value - depenseReve, 0);
    await this.update({ "data.reve.reve": myReve });
    
    if (myReve.value == 0) { // 0 points de reve
      ChatMessage.create({ content: this.name + " est réduit à 0 Points de Rêve, et tombe endormi !" });
      closeTMR = true;
    }
    if (closeTMR) {
      this.currentTMR.close(); // Close TMR !
    } else {
      this.currentTMR.maximize(); // Re-display TMR
    }
    // Final chat message
    let chatOptions = {
      content: "<strong>Test : " + rollData.selectedCarac.label + " / " + rollData.selectedDraconic.name + " / " + rollData.selectedSort.name + "</strong>"
      + RdDResolutionTable.explainRollData(rollData)
      + explications
    }
    
    ChatUtility.chatWithRollMode(chatOptions, this.name)
  }

  /* -------------------------------------------- */  
  async rollCarac( caracName ) {
    let rollData = {
      selectedCarac: this.getCaracByName(caracName),
      needSignificative : !this.isEntiteCauchemar() && this.data.data.sante.sonne.value
    };

    const dialog = await RdDRoll.create(this, rollData,
      {html: 'systems/foundryvtt-reve-de-dragon/templates/dialog-roll-carac.html'},
      {
        name: 'jet-'+caracName,
        label: 'Jet ' + Grammar.apostrophe('de', rollData.selectedCarac.label),
        callbacks: [
          this.createCallbackExperience(),
          { action: this._rollCaracResult }
        ]
      }
    );
    dialog.render(true);
  }

  /* -------------------------------------------- */  
  async _rollCaracResult(rollData) {
    rollData.show = {
      title: rollData.selectedCarac.label,
      points: true
    };
    // Final chat message
    let chatOptions = { content: await RdDResolutionTable.explainRollDataV2(rollData) }
    
    ChatUtility.chatWithRollMode(chatOptions, this.name)
  }

  /* -------------------------------------------- */  
  async rollCompetence( name ) {
    let rollData = {
      competence:  duplicate(this.getCompetence(name)),
      needSignificative : !this.isEntiteCauchemar() && this.data.data.sante.sonne.value
    }
    
    if (rollData.competence.type == 'competencecreature') {
      // Fake competence pour créature
      rollData.competence.data = { defaut_carac: "carac_creature",  categorie: "creature" };
      rollData.carac = { carac_creature: { label: competence.name, value: competence.data.carac_value } };
    }
    else{
      rollData.carac = this.data.data.carac;
    }
    console.log("rollCompetence !!!", rollData);

    const dialog = await RdDRoll.create(this, rollData, {html: 'systems/foundryvtt-reve-de-dragon/templates/dialog-competence.html'}, {
        name: 'jet-competence',
        label: 'Jet ' +Grammar.apostrophe('de', name),
        callbacks: [
          this.createCallbackExperience(),
          { action: this._competenceResult }
        ]
      } );
    dialog.render(true);    
  }

  /* -------------------------------------------- */
  async creerTacheDepuisLivre( item ) {
    let tache = { name: "Lire " + item.name, type: 'tache', 
                  data: { 
                    carac: 'intellect',
                    competence: 'Ecriture',
                    difficulte: item.data.data.difficulte,
                    periodicite: "60 minutes",
                    fatigue: 2,
                    points_de_tache: item.data.data.points_de_tache,
                    points_de_tache_courant: 0,
                    description: "Lecture du livre " + item.name +
                                  " - XP : " + item.data.data.xp + " - Compétences : " + item.data.data.competence
                  }
                }
    await this.createOwnedItem( tache, { renderSheet: true } );
  }
  
  /* -------------------------------------------- */
  getTache ( id )  {
    return this.data.items.find( item => item._id == id );
  }

  /* -------------------------------------------- */
  async rollTache( id ) {
    let tache = duplicate( this.getTache( id ) );
    let competence = duplicate(this.getCompetence(tache.data.competence));
    competence.data.defaut_carac = tache.data.carac; // Patch !
    let rollData = {
      competence:  competence,
      needSignificative : this.data.data.sante.sonne.value,
      tache: tache,
      diffConditions: tache.data.difficulte,
      editLibre: false,
      editConditions: false,
      actor: this
    }

    rollData.carac = {};
    rollData.carac[tache.data.carac] = duplicate(this.data.data.carac[tache.data.carac]); // Single carac
    console.log("rollTache !!!", duplicate(rollData));

    const dialog = await RdDRoll.create(this, rollData, {html: 'systems/foundryvtt-reve-de-dragon/templates/dialog-competence.html'}, {
        name: 'jet-competence',
        label: 'Jet de Tâche ' + tache.name,
        callbacks: [
          this.createCallbackExperience(),
          { condition: r=> r.rolled.isETotal, action: r => this._tacheETotal(r)},
          { action: r => this._tacheResult(r) }
        ]
      } );
    dialog.render(true);
  }

  /* -------------------------------------------- */
  _tacheResult(rollData) {
    // Mise à jour de la tache
    rollData.tache.data.points_de_tache_courant += rollData.rolled.ptTache;
    this.updateEmbeddedEntity( "OwnedItem", rollData.tache);
    this.santeIncDec( "fatigue", rollData.tache.data.fatigue);

    // Message de résultat
    ChatUtility.chatWithRollMode({
      content: "<strong>Test de Tache : " + rollData.tache.name + " - " + rollData.selectedCarac.label + " / " + rollData.competence.name + "</strong>"
      + RdDResolutionTable.explainRollData(rollData)
      + "<br>Points de taches : " + rollData.rolled.ptTache + ", ajustement qualité: " + rollData.rolled.ptQualite      
    }, this.name);
    // Message spécifique de la tâche
    ChatUtility.chatWithRollMode({
      content: "Votre tâche <strong>" + rollData.tache.name + "</strong> a duré " + rollData.tache.data.periodicite + "."
      + "<br>Votre avancement est désormais de " + rollData.tache.data.points_de_tache_courant + " Points de Tache sur un objectif de "
      + rollData.tache.data.points_de_tache + "."
      + "<br>Et vous vous êtes fatigué de " +       rollData.tache.data.fatigue  + " cases."
    }, this.name);
  }

  /* -------------------------------------------- */
  _tacheETotal(rollData) {
    rollData.tache.data.difficulte--;
    this.updateEmbeddedEntity( "OwnedItem", rollData.tache);
    ChatUtility.chatWithRollMode( {
      content: "Vous avez fait un Echec Total sur votre Tache <strong>" + rollData.tache.name + "</strong> : en conséquence, la difficulté a augmenté de 1"
    }, this.name);
  }

  /* -------------------------------------------- */
  _competenceResult(rollData) {
    ChatUtility.chatWithRollMode({
      content: "<strong>Test : " + rollData.selectedCarac.label + " / " + rollData.competence.name + "</strong>"
      + RdDResolutionTable.explainRollData(rollData)
      + "<br>Points de taches : " + rollData.rolled.ptTache + ", ajustement qualité: " + rollData.rolled.ptQualite
    }, this.name);    
  }

  /* -------------------------------------------- */
  async rollAppelChance( )
  {
    let rollData = { selectedCarac: this.getCaracByName('chance-actuelle') };

    const dialog = await RdDRoll.create(this, rollData,
      { html: 'systems/foundryvtt-reve-de-dragon/templates/dialog-roll-carac.html'},
      {
        name: 'appelChance',
        label: 'Appel à la chance',
        callbacks: [
          this.createCallbackExperience(),
          { action: r => this._appelChanceResult(r) }
        ]
      }
    );
    dialog.render(true);
  }

  /* -------------------------------------------- */
  async _appelChanceResult(rollData) {
    const message = {
      user: game.user._id,
      alias: this.name,
      content: this.name + " fait appel à la chance" + RdDResolutionTable.explain(rollData.rolled)
    };
    if (rollData.rolled.isSuccess) {
      message.content += "<br>Un point de chance est dépensée, l'action peut être retentée"
      await this.chanceActuelleIncDec(-1)
    }
    ChatMessage.create(message);
  }
  
  /* -------------------------------------------- */
  async chanceActuelleIncDec(value) {
    let chance = duplicate(this.data.data.compteurs.chance);
    chance.value = Math.max(chance.value + value, 0);
    await this.update( {"data.compteurs.chance": chance } );
  }

  /* -------------------------------------------- */
  ajustementAstrologique() {
    if (this.isCreature()) {
      return 0;
    }
    // selon l'heure de naissance...
    return game.system.rdd.calendrier.getAjustementAstrologique(this.data.data.heure, this.data.name);
  }
  /* -------------------------------------------- */
  checkDesirLancinant( ) {
    let queue = this.data.items.filter( (item) => item.name.toLowerCase().includes('lancinant') );
    return (queue.length > 0);
  }

  /* -------------------------------------------- */
  async appliquerExperience( rolled, caracName, competenceName = undefined ) {

    if ( rolled.isPart && rolled.finalLevel < 0) {
      // Cas de désir lancinant, pas d'expérience sur particulière
      if ( this.checkDesirLancinant() ) {
        ChatMessage.create( { content: `Vous souffrez au moins d'un Désir Lancinant, vous ne pouvez pas gagner d'expérience sur une Particulière tant que le désir n'est pas assouvi`, 
                      whisper: ChatMessage.getWhisperRecipients(game.user.name) } );
        return { result:false, xpcarac:0, xpCompetence: 0 };
      }

      if (caracName == 'derobee') caracName = 'agilite';
      let carac = duplicate(this.data.data.carac);

      let xp = Math.abs(rolled.finalLevel);
      let xpCarac = Math.floor(xp / 2); // impair: arrondi inférieur en carac

      let xpComp = 0;
      if ( competenceName  ) {
        xpComp = xp - xpCarac;
        let competence = duplicate( RdDUtility.findCompetence( this.data.items, competenceName ) );
        competence.data.xp += xpComp;
        await this.updateEmbeddedEntity( "OwnedItem", competence);
      } else {
        xpCarac = Math.max(xpCarac, 1);
      } 
      if ( !carac[caracName].isderivee) {
        carac[caracName].xp += xpCarac;
        await this.update( {"data.carac": carac } );
      } else {
        ChatMessage.create( { content: `Vous avez ${xpCarac} à répartir pour la caractérisque dérivée ${caracName}. Vous devez le faire manuellement.`, 
                              whisper: ChatMessage.getWhisperRecipients(game.user.name) } );
      }
      return { result:true, xpcarac:xpCarac, xpCompetence: xpComp }; //XP 
    }

    return { result:false, xpcarac:0, xpCompetence: 0 }; // Pas d'XP
  }
  
  /* -------------------------------------------- */
  async ajouteNombreAstral( data ) {
    // Gestion expérience (si existante)
    this.appliquerExperience( data.rolled, "vue", "astrologie");

    // Ajout du nombre astral
    const item = {name: "Nombre Astral", type: "nombreastral", data: 
              { value: data.nbAstral, istrue: data.isvalid, jourindex: Number(data.date), jourlabel: game.system.rdd.calendrier.getDateFromIndex( Number(data.date) )  } };
    await this.createEmbeddedEntity("OwnedItem", item);
    
    // Suppression des anciens nombres astraux
    let toDelete = this.data.items.filter( (item) => item.data.jourindex < game.system.rdd.calendrier.getCurrentDayIndex() );
    const deletions = toDelete.map(i => i._id);
    await this.deleteEmbeddedEntity("OwnedItem", deletions);

    // Affichage Dialog
    this.astrologieNombresAstraux();
  } 

  /* -------------------------------------------- */
  async astrologieNombresAstraux( ) { 
    // Afficher l'interface spéciale
    const astrologieDialog = await RdDAstrologieJoueur.create( this, {} );
    astrologieDialog.render(true);
  }

  /* -------------------------------------------- */
  getCaracByName(caracName) {
    switch (caracName)
    {
      case 'reve-actuel':
        return {
          label: 'Rêve Actuel',
          value: this.getReveActuel(),
          type: "number",
          ignoreEtatGeneral: true
        };
      case 'chance-actuelle':
        return {
          type: "number",
          value: this.getChanceActuel(),
          label: 'Chance actuelle',
          ignoreEtatGeneral: true
        };
      default:
        return this.data.data.carac[caracName]; // Per default
    }
  }

  /* -------------------------------------------- */
  getSortList() {
    return this.data.items.filter(item => item.type == "sort");
  }
  
  /* -------------------------------------------- */
  getDraconicList() {
    return this.data.items.filter(item => item.data.categorie == 'draconic')
  }
  
  /* -------------------------------------------- */  
  checkMonteeLaborieuse( ) { // Return +1 si la queue Montée Laborieuse est présente, 0 sinon
    let monteLaborieuse = this.data.items.find( item => (item.type == 'queue' || item.type == 'souffle') && item.name.toLowerCase().includes( 'montée laborieuse' ) );
    if ( monteLaborieuse ) {
      ChatMessage.create( {
        content: "Vous êtes sous le coup d'une Montée Laborieuse : vos montées en TMR coûtent 1 Point de Rêve de plus.",
        whisper: ChatMessage.getWhisperRecipients(game.user.name) } );
      return 1;
    }
    return 0;
  }

  /* -------------------------------------------- */  
  async displayTMR(mode="normal") 
  {
    let isRapide= mode == "rapide"; 
    if (mode != "visu")
    {
      let minReveValue = (isRapide) ? 3 : 2;
      if (this.getReveActuel() < minReveValue ) {
        ChatMessage.create( {
          content: "Vous n'avez plus assez de Points de Reve pour monter dans les Terres Médianes",
          whisper: ChatMessage.getWhisperRecipients(game.user.name) } );
        return;
      } 
    }
    
    // Notification au MJ
    ChatMessage.create( { content: game.user.name + " est monté dans les TMR en mode : " + mode, whisper: ChatMessage.getWhisperRecipients("GM") } );

    let data = {
      fatigue: {
        malus: RdDUtility.calculMalusFatigue(this.data.data.sante.fatigue.value,  this.data.data.sante.endurance.max),
        html: "<table class='table-fatigue'>" + RdDUtility.makeHTMLfatigueMatrix( this.data.data.sante.fatigue.value,  this.data.data.sante.endurance.max ).html() + "</table>"
      },
      draconic: this.getDraconicList(),
      sort: this.getSortList(),
      caracReve: this.data.data.carac.reve.value,
      pointsReve: this.getReveActuel(),
      isRapide: isRapide
    }    
    let html = await renderTemplate('systems/foundryvtt-reve-de-dragon/templates/dialog-tmr.html', data );
    this.currentTMR = new RdDTMRDialog(html, this, data, mode);
    this.currentTMR.render(true);
  }

  
  /* -------------------------------------------- */  
  async rollCompetenceCreature( compName ) {
    let competence = this.getCompetence(compName);
    if ( competence.type == 'competencecreature' && competence.data.iscombat ) {
      let competence = this.getCompetence(compName);
      RdDCombat.createUsingTarget(this).attaque(competence, armeItem);
    }
    else {
      this.rollCompetence(name);
    }
  }

  /* -------------------------------------------- */  
  rollArme(compName,  armeName = undefined) {

    let armeItem = this.data.items.find(item=>item.type==="arme" && (item.name === armeName));
    if (armeItem && compName == undefined) compName = armeItem.data.competence;
    let competence = this.getCompetence(compName == undefined? armeName : compName);
    
    if (armeItem || armeName || (competence.type == 'competencecreature' && competence.data.iscombat)) {
      RdDCombat.createUsingTarget(this).attaque(competence, armeItem);
    } else {
      this.rollCompetence( competence.name );
    }
  }

  /* -------------------------------------------- */
  _getTarget() {
    if (game.user.targets && game.user.targets.size == 1) {
      for (let target of game.user.targets) {
        return target;
      }
    }
    return undefined;
  }
  
  /* -------------------------------------------- */  
  async equiperObjet( itemID ) {
    let item = this.getOwnedItem(itemID);
    if ( item && item.data.data ) {
      let update = {_id: item._id, "data.equipe": !item.data.data.equipe };
      await this.updateEmbeddedEntity("OwnedItem", update); 
      this.computeEncombrementTotalEtMalusArmure(); // Mise à jour encombrement
    }
  }
  
  /* -------------------------------------------- */  
  computeArmure( attackerRoll ) {
    let locData = attackerRoll.loc;
    let dmg = attackerRoll.dmg.dmgArme + attackerRoll.dmg.dmgActor;
    let arme =  attackerRoll.arme;
    // TODO: arc ignore armure
    let protection = 0;
    for (const item of this.data.items) {
      if (item.type == "armure" && item.data.equipe) {
        let update = duplicate(item);
        protection += new Roll(update.data.protection.toString()).roll().total;
        update.data.deterioration = Misc.toInt(update.data.deterioration) + dmg;
        dmg = 0; // Reset it
        if ( update.data.deterioration >= 10) {          
          update.data.deterioration = 0;
          let res = /\d+^/.exec(update.data.protection);
          if ( res ) 
            update.data.protection = "1d"+update.data.protection;
          // if ( update.data.protection.toString().length == 1 ) 
          // update.data.protection = "1d"+update.data.protection;
          else if (res = /(\d+d\d+)(\-\d+)?/.exec( update.data.protection ))
          {
            let malus = Misc.toInt(res[2]) - 1;
            update.data.protection = res[1] + malus;
          }
          else{
            ui.notifications.warn(`La valeur d'armure de votre ${item.name} est incorrecte`)
          }
          ChatMessage.create( {content: "Détérioration d'armure: " + update.data.protection } );
        }
        this.updateEmbeddedEntity("OwnedItem", update); 
      }
    }
    const penetration = arme ? Misc.toInt(arme.data.penetration) : 0;
    protection = Math.max(protection - penetration, 0);
    protection += this.getProtectionNaturelle();
    console.log("Final protect", protection);
    return protection;
  }

    /* -------------------------------------------- */
    async encaisser( ) {
      let data = { ajustementsEncaissement: RdDUtility.getAjustementsEncaissement() };
      let html = await renderTemplate('systems/foundryvtt-reve-de-dragon/templates/dialog-roll-encaisser.html', data );
      new RdDEncaisser(html, this ).render(true);
    }
  
  /* -------------------------------------------- */
  async encaisser( ) {
    let data = { ajustementsEncaissement: RdDUtility.getAjustementsEncaissement() };
    let html = await renderTemplate('systems/foundryvtt-reve-de-dragon/templates/dialog-roll-encaisser.html', data );
    new RdDEncaisser(html, this).render(true);
  }

  /* -------------------------------------------- */
  async encaisserDommages( attackerRoll, attacker = undefined ) {
    if (attacker && !await attacker.accorder(this, 'avant-encaissement')) {
      return;
    }

    console.log("encaisserDommages", attackerRoll )
    const armure = this.computeArmure( attackerRoll );

    const rollEncaissement = new Roll("2d10 + @dmg - @armure",{
      dmg: attackerRoll.dmg.total,
      armure: armure
    }).roll();
    RdDDice.show(rollEncaissement);
    let result = RdDUtility.computeBlessuresSante(rollEncaissement.total, attackerRoll.dmg.mortalite, attackerRoll.dmg.loc);
    result.endurance = Math.max(result.endurance, -Number(this.data.data.sante.endurance.value));
    await this.santeIncDec("vie", result.vie);
    await this.santeIncDec("endurance", result.endurance, (result.critiques > 0));

    this.manageBlessures(result); // Will upate the result table
    const blessureLegere = (result.legeres > 0 ? "une blessure légère" : "");
    const blessureGrave = (result.graves > 0 ? "une blessure grave" : "");
    const blessureCritique = (result.critiques > 0 ? "une blessure critique" : "");
    let commonMsg = {
       title: "Blessures !", 
       roll: rollEncaissement ,
       content: this.data.name + " a encaissé " + blessureLegere + blessureGrave + blessureCritique
    }
    let addonMsg = "<br>Et a perdu : <br>" + result.endurance + " Endurance et " + result.vie + " Points de Vie";
    if ( this.hasPlayerOwner ) {
      commonMsg.content += addonMsg; // Message pour tout le monde
      ChatMessage.create( commonMsg );
    } else { // Le defenseur n'est pas un PJ, donc message complet uniquement pour le MJ
      ChatMessage.create( commonMsg ); // Message pour tout le monde
      let gmMsg = duplicate(commonMsg);
      gmMsg.content = addonMsg; // Et message complémentaire uniquement pour le MJ
      gmMsg.whisper = ChatMessage.getWhisperRecipients( "GM" );
      ChatMessage.create( gmMsg );
    }

    this.computeEtatGeneral();
    this.sheet.render(true);
  }

  /* -------------------------------------------- */  
  /** @override */
  getRollData() {
    const data = super.getRollData();
    return data;
  }
  
  /* -------------------------------------------- */  
  /* -- entites -- */
  /* retourne true si on peut continuer, false si on ne peut pas continuer */
  async targetEntiteNonAccordee(target, when='avant-encaissement')
  {
    if (target)
    {
      return !await this.accorder(target.actor, when);
    }
    return false;
  }

  /* -------------------------------------------- */  
  async accorder(entite, when = 'avant-encaissement')
  {
    if (when != game.settings.get("foundryvtt-reve-de-dragon", "accorder-entite-cauchemar")
     || !entite.isEntiteCauchemar()
     || entite.isEntiteCauchemarAccordee(this)) {
      return true;
    }
    
    let rolled = await RdDResolutionTable.roll( this.getReveActuel(), - Number(entite.data.data.carac.niveau.value));

    let message = {
      content: "Jet de points actuels de rêve à " + rolled.finalLevel + RdDResolutionTable.explain(rolled) + "<br>",
      whisper: ChatMessage.getWhisperRecipients(this.name)
    };

    if (rolled.isSuccess) {
      await entite.setEntiteReveAccordee(this);
      message.content += this.name + " s'est accordé avec " + entite.name;
    }
    else {
      message.content+= this.name + " n'est pas accordé avec " + entite.name;
    }

    ChatMessage.create( message );
    return rolled.isSuccess;
  }

  /* -------------------------------------------- */  
  isEntiteCauchemar()
  {
    return this.data.type == 'entite';
  }

  /* -------------------------------------------- */  
  isEntiteCauchemarAccordee(attaquant)
  {
    if (!this.isEntiteCauchemar()) { return true; }
    let resonnance = this.data.data.sante.resonnance;
    return (resonnance.actors.find(it => it == attaquant._id));
  }

  /* -------------------------------------------- */  
  async setEntiteReveAccordee(attaquant)
  {
    if (!this.isEntiteCauchemar()) {
      ui.notifications.error("Impossible de s'accorder à " + this.name + ": ce n'est pas une entite de cauchemer/rêve");
      return;
    }
    let resonnance = duplicate(this.data.data.sante.resonnance);
    if (resonnance.actors.find(it => it == attaquant._id)){
      // déjà accordé
      return;
    }
    resonnance.actors.push(attaquant._id);
    await this.update( {"data.sante.resonnance": resonnance});
    return;
  }
  /* -------------------------------------------- */ 
  async optimizeArgent( sumDenier ) {
    let sols = Math.floor(sumDenier / 100);
    let deniers = sumDenier - (sols*100);
    let nbOr = Math.floor(sols / 10);
    let nbArgent = sols - (nbOr*10);
    let nbBronze = Math.floor(deniers / 10 );
    let nbEtain = deniers - (nbBronze*10);

    //    console.log("ARGENT", nbOr, nbArgent, nbBronze, nbEtain);
    let piece = this.data.items.find( item => item.type =='monnaie' && item.data.valeur_deniers == 1000);
    if (piece ) {
      let update = { _id: piece._id, 'data.quantite': nbOr};
      const updated = await this.updateEmbeddedEntity("OwnedItem", update); 
    }

    piece = this.data.items.find( item => item.type =='monnaie' && item.data.valeur_deniers == 100);
    if (piece ) {
      let update = { _id: piece._id, 'data.quantite': nbArgent};
      const updated = await this.updateEmbeddedEntity("OwnedItem", update); 
    }

    piece = this.data.items.find( item => item.type =='monnaie' && item.data.valeur_deniers == 10);
    if (piece ) {
      let update = { _id: piece._id, 'data.quantite': nbBronze};
      const updated = await this.updateEmbeddedEntity("OwnedItem", update); 
    }

    piece = this.data.items.find( item => item.type =='monnaie' && item.data.valeur_deniers == 1);
    if (piece ) {
      let update = { _id: piece._id, 'data.quantite': nbEtain};
      const updated = await this.updateEmbeddedEntity("OwnedItem", update); 
    }
  }

  /* -------------------------------------------- */ 
  payerDenier( sumDenier ) {
    sumDenier = Number(sumDenier);
    let denierDisponible = 0;
    let monnaie = this.data.items.filter( item => item.type =='monnaie');
    for ( let piece of monnaie ) {
      denierDisponible += piece.data.valeur_deniers * Number(piece.data.quantite);      
    }
    console.log("DENIER", game.user.character, sumDenier, denierDisponible);
    if ( denierDisponible >= sumDenier) {
      denierDisponible -= sumDenier;
      this.optimizeArgent(denierDisponible);
    } else {
      let message = { 
        whisper: ChatUtility.getWhisperRecipientsAndGMs( this.name ),
        content : "Vous n'avez pas assez d'argent pour paye cette somme !"
      };
      ChatMessage.create( message );
    }
  }

  /* -------------------------------------------- */ 
  async monnaieIncDec( id, value) {
    let monnaie = this.data.items.find( item => item.type =='monnaie' && item._id == id );
    if (monnaie ) {
      monnaie.data.quantite += value;
      if ( monnaie.data.quantite < 0 ) monnaie.data.quantite = 0; // Sanity check
      const update = {_id: monnaie._id, 'data.quantite': monnaie.data.quantite };
      const updated = await this.updateEmbeddedEntity("OwnedItem", update); 
    }
  }

}


